@@ -3,12 +3,10 @@ use std::{convert::TryInto, fmt, num::TryFromIntError, sync::Arc, time::Duration
 use rand::RngCore;
 use thiserror::Error;
 
-#[cfg(feature = "rustls")]
-use crate::crypto::types::{Certificate, CertificateChain, PrivateKey};
 use crate::{
     cid_generator::{ConnectionIdGenerator, RandomConnectionIdGenerator},
     congestion,
-    crypto::{self, ClientConfig as _, HandshakeTokenKey as _, HmacKey as _, ServerConfig as _},
+    crypto::{self, HandshakeTokenKey, HmacKey},
     VarInt, VarIntBoundsExceeded, DEFAULT_SUPPORTED_VERSIONS,
 };
 
@@ -27,7 +25,7 @@ use crate::{
 pub struct TransportConfig {
     pub(crate) max_concurrent_bidi_streams: VarInt,
     pub(crate) max_concurrent_uni_streams: VarInt,
-    pub(crate) max_idle_timeout: Option<Duration>,
+    pub(crate) max_idle_timeout: Option<VarInt>,
     pub(crate) stream_receive_window: VarInt,
     pub(crate) receive_window: VarInt,
     pub(crate) send_window: u64,
@@ -48,21 +46,21 @@ pub struct TransportConfig {
 }
 
 impl TransportConfig {
-    /// Maximum number of bidirectional streams that may be open concurrently
+    /// Maximum number of incoming bidirectional streams that may be open concurrently
     ///
     /// Must be nonzero for the peer to open any bidirectional streams.
     ///
     /// Worst-case memory use is directly proportional to `max_concurrent_bidi_streams *
     /// stream_receive_window`, with an upper bound proportional to `receive_window`.
-    pub fn max_concurrent_bidi_streams(&mut self, value: u64) -> Result<&mut Self, ConfigError> {
-        self.max_concurrent_bidi_streams = value.try_into()?;
-        Ok(self)
+    pub fn max_concurrent_bidi_streams(&mut self, value: VarInt) -> &mut Self {
+        self.max_concurrent_bidi_streams = value;
+        self
     }
 
     /// Variant of `max_concurrent_bidi_streams` affecting unidirectional streams
-    pub fn max_concurrent_uni_streams(&mut self, value: u64) -> Result<&mut Self, ConfigError> {
-        self.max_concurrent_uni_streams = value.try_into()?;
-        Ok(self)
+    pub fn max_concurrent_uni_streams(&mut self, value: VarInt) -> &mut Self {
+        self.max_concurrent_uni_streams = value;
+        self
     }
 
     /// Maximum duration of inactivity to accept before timing out the connection.
@@ -72,12 +70,24 @@ impl TransportConfig {
     ///
     /// **WARNING**: If a peer or its network path malfunctions or acts maliciously, an infinite
     /// idle timeout can result in permanently hung futures!
-    pub fn max_idle_timeout(&mut self, value: Option<Duration>) -> Result<&mut Self, ConfigError> {
-        if value.map_or(false, |x| x.as_millis() > VarInt::MAX.0 as u128) {
-            return Err(ConfigError::OutOfBounds);
-        }
-        self.max_idle_timeout = value;
-        Ok(self)
+    ///
+    /// ```
+    /// # use std::{convert::TryInto, time::Duration};
+    /// # use quinn_proto::{TransportConfig, VarInt, VarIntBoundsExceeded};
+    /// # fn main() -> Result<(), VarIntBoundsExceeded> {
+    /// let mut config = TransportConfig::default();
+    ///
+    /// // Set the idle timeout as `VarInt`-encoded milliseconds
+    /// config.max_idle_timeout(Some(VarInt::from_u32(10_000).into()));
+    ///
+    /// // Set the idle timeout as a `Duration`
+    /// config.max_idle_timeout(Some(Duration::from_secs(10).try_into()?));
+    /// # Ok(())
+    /// # }
+    /// ```
+    pub fn max_idle_timeout(&mut self, value: Option<IdleTimeout>) -> &mut Self {
+        self.max_idle_timeout = value.map(|t| t.0);
+        self
     }
 
     /// Maximum number of bytes the peer may transmit without acknowledgement on any one stream
@@ -88,9 +98,9 @@ impl TransportConfig {
     /// stream doesn't monopolize receive buffers, which may otherwise occur if the application
     /// chooses not to read from a large stream for a time while still requiring data on other
     /// streams.
-    pub fn stream_receive_window(&mut self, value: u64) -> Result<&mut Self, ConfigError> {
-        self.stream_receive_window = value.try_into()?;
-        Ok(self)
+    pub fn stream_receive_window(&mut self, value: VarInt) -> &mut Self {
+        self.stream_receive_window = value;
+        self
     }
 
     /// Maximum number of bytes the peer may transmit across all streams of a connection before
@@ -99,9 +109,9 @@ impl TransportConfig {
     /// This should be set to at least the expected connection latency multiplied by the maximum
     /// desired throughput. Larger values can be useful to allow maximum throughput within a
     /// stream while another is blocked.
-    pub fn receive_window(&mut self, value: u64) -> Result<&mut Self, ConfigError> {
-        self.receive_window = value.try_into()?;
-        Ok(self)
+    pub fn receive_window(&mut self, value: VarInt) -> &mut Self {
+        self.receive_window = value;
+        self
     }
 
     /// Maximum number of bytes to transmit to a peer without acknowledgment
@@ -227,7 +237,7 @@ impl Default for TransportConfig {
         TransportConfig {
             max_concurrent_bidi_streams: 100u32.into(),
             max_concurrent_uni_streams: 100u32.into(),
-            max_idle_timeout: Some(Duration::from_millis(10_000)),
+            max_idle_timeout: Some(VarInt(10_000)),
             stream_receive_window: STREAM_RWND.into(),
             receive_window: VarInt::MAX,
             send_window: (8 * STREAM_RWND).into(),
@@ -244,7 +254,7 @@ impl Default for TransportConfig {
             datagram_receive_buffer_size: Some(STREAM_RWND as usize),
             datagram_send_buffer_size: 1024 * 1024,
 
-            congestion_controller_factory: Box::new(Arc::new(congestion::NewRenoConfig::default())),
+            congestion_controller_factory: Box::new(Arc::new(congestion::CubicConfig::default())),
         }
     }
 }
@@ -288,11 +298,9 @@ impl fmt::Debug for TransportConfig {
 /// Global configuration for the endpoint, affecting all connections
 ///
 /// Default values should be suitable for most internet applications.
-pub struct EndpointConfig<S>
-where
-    S: crypto::Session,
-{
-    pub(crate) reset_key: Arc<S::HmacKey>,
+#[derive(Clone)]
+pub struct EndpointConfig {
+    pub(crate) reset_key: Arc<dyn HmacKey>,
     pub(crate) max_udp_payload_size: VarInt,
     /// CID generator factory
     ///
@@ -300,22 +308,17 @@ where
     pub(crate) connection_id_generator_factory:
         Arc<dyn Fn() -> Box<dyn ConnectionIdGenerator> + Send + Sync>,
     pub(crate) supported_versions: Vec<u32>,
-    pub(crate) initial_version: u32,
 }
 
-impl<S> EndpointConfig<S>
-where
-    S: crypto::Session,
-{
+impl EndpointConfig {
     /// Create a default config with a particular `reset_key`
-    pub fn new(reset_key: S::HmacKey) -> Self {
+    pub fn new(reset_key: Arc<dyn HmacKey>) -> Self {
         let cid_factory: fn() -> Box<dyn ConnectionIdGenerator> =
             || Box::new(RandomConnectionIdGenerator::default());
         Self {
-            reset_key: Arc::new(reset_key),
+            reset_key,
             max_udp_payload_size: 1480u32.into(), // Typical internet MTU minus IPv4 and UDP overhead, rounded up to a multiple of 8
             connection_id_generator_factory: Arc::new(cid_factory),
-            initial_version: DEFAULT_SUPPORTED_VERSIONS[0],
             supported_versions: DEFAULT_SUPPORTED_VERSIONS.to_vec(),
         }
     }
@@ -340,9 +343,9 @@ where
 
     /// Private key used to send authenticated connection resets to peers who were
     /// communicating with a previous instance of this endpoint.
-    pub fn reset_key(&mut self, value: &[u8]) -> Result<&mut Self, ConfigError> {
-        self.reset_key = Arc::new(S::HmacKey::new(value)?);
-        Ok(self)
+    pub fn reset_key(&mut self, key: Arc<dyn HmacKey>) -> &mut Self {
+        self.reset_key = key;
+        self
     }
 
     /// Maximum UDP payload size accepted from peers. Excludes UDP and IP overhead.
@@ -368,79 +371,58 @@ where
     }
 
     /// Override supported QUIC versions
-    pub fn supported_versions(
-        &mut self,
-        supported_versions: Vec<u32>,
-        initial_version: u32,
-    ) -> Result<&mut Self, ConfigError> {
-        if !supported_versions.contains(&initial_version) {
-            return Err(ConfigError::OutOfBounds);
-        }
+    pub fn supported_versions(&mut self, supported_versions: Vec<u32>) -> &mut Self {
         self.supported_versions = supported_versions;
-        self.initial_version = initial_version;
-        Ok(self)
+        self
     }
 }
 
-impl<S: crypto::Session> fmt::Debug for EndpointConfig<S> {
+impl fmt::Debug for EndpointConfig {
     fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
         fmt.debug_struct("EndpointConfig")
             .field("reset_key", &"[ elided ]")
             .field("max_udp_payload_size", &self.max_udp_payload_size)
             .field("cid_generator_factory", &"[ elided ]")
             .field("supported_versions", &self.supported_versions)
-            .field("initial_version", &self.initial_version)
             .finish()
     }
 }
 
-impl<S: crypto::Session> Default for EndpointConfig<S> {
+#[cfg(feature = "ring")]
+impl Default for EndpointConfig {
     fn default() -> Self {
-        let mut reset_key = vec![0; S::HmacKey::KEY_LEN];
+        let mut reset_key = [0; 64];
         rand::thread_rng().fill_bytes(&mut reset_key);
-        Self::new(
-            S::HmacKey::new(&reset_key)
-                .expect("HMAC key rejected random bytes; use EndpointConfig::new instead"),
-        )
-    }
-}
 
-impl<S: crypto::Session> Clone for EndpointConfig<S> {
-    fn clone(&self) -> Self {
-        Self {
-            reset_key: self.reset_key.clone(),
-            max_udp_payload_size: self.max_udp_payload_size,
-            connection_id_generator_factory: self.connection_id_generator_factory.clone(),
-            supported_versions: self.supported_versions.clone(),
-            initial_version: self.initial_version,
-        }
+        Self::new(Arc::new(ring::hmac::Key::new(
+            ring::hmac::HMAC_SHA256,
+            &reset_key,
+        )))
     }
 }
 
 /// Parameters governing incoming connections
 ///
 /// Default values should be suitable for most internet applications.
-pub struct ServerConfig<S>
-where
-    S: crypto::Session,
-{
+#[derive(Clone)]
+pub struct ServerConfig {
     /// Transport configuration to use for incoming connections
     pub transport: Arc<TransportConfig>,
 
     /// TLS configuration used for incoming connections.
     ///
     /// Must be set to use TLS 1.3 only.
-    pub crypto: S::ServerConfig,
+    pub crypto: Arc<dyn crypto::ServerConfig>,
 
     /// Used to generate one-time AEAD keys to protect handshake tokens
-    pub(crate) token_key: Arc<S::HandshakeTokenKey>,
+    pub(crate) token_key: Arc<dyn HandshakeTokenKey>,
 
     /// Whether to require clients to prove ownership of an address before committing resources.
     ///
     /// Introduces an additional round-trip to the handshake to make denial of service attacks more difficult.
-    pub(crate) use_stateless_retry: bool,
+    pub(crate) use_retry: bool,
     /// Microseconds after a stateless retry token was issued for which it's considered valid.
-    pub(crate) retry_token_lifetime: u64,
+    pub(crate) retry_token_lifetime: Duration,
 
     /// Maximum number of concurrent connections
     pub(crate) concurrent_connections: u32,
@@ -452,19 +434,19 @@ where
     pub(crate) migration: bool,
 }
 
-impl<S> ServerConfig<S>
-where
-    S: crypto::Session,
-{
-    /// Create a default config with a particular `master_key`
-    pub fn new(prk: S::HandshakeTokenKey) -> Self {
+impl ServerConfig {
+    /// Create a default config with a particular handshake token key
+    pub fn new(
+        crypto: Arc<dyn crypto::ServerConfig>,
+        token_key: Arc<dyn HandshakeTokenKey>,
+    ) -> Self {
         Self {
             transport: Arc::new(TransportConfig::default()),
-            crypto: S::ServerConfig::new(),
+            crypto,
 
-            token_key: Arc::new(prk),
-            use_stateless_retry: false,
-            retry_token_lifetime: 15_000_000,
+            token_key,
+            use_retry: false,
+            retry_token_lifetime: Duration::from_secs(15),
 
             concurrent_connections: 100_000,
 
@@ -473,28 +455,32 @@ where
     }
 
     /// Private key used to authenticate data included in handshake tokens.
-    pub fn token_key(&mut self, master_key: &[u8]) -> Result<&mut Self, ConfigError> {
-        self.token_key = Arc::new(S::HandshakeTokenKey::from_secret(&master_key));
+    pub fn token_key(
+        &mut self,
+        value: Arc<dyn HandshakeTokenKey>,
+    ) -> Result<&mut Self, ConfigError> {
+        self.token_key = value;
         Ok(self)
     }
 
     /// Whether to require clients to prove ownership of an address before committing resources.
     ///
     /// Introduces an additional round-trip to the handshake to make denial of service attacks more difficult.
-    pub fn use_stateless_retry(&mut self, value: bool) -> &mut Self {
-        self.use_stateless_retry = value;
+    pub fn use_retry(&mut self, value: bool) -> &mut Self {
+        self.use_retry = value;
         self
     }
 
-    /// Microseconds after a stateless retry token was issued for which it's considered valid.
-    pub fn retry_token_lifetime(&mut self, value: u64) -> &mut Self {
+    /// Duration after a stateless retry token was issued for which it's considered valid.
+    pub fn retry_token_lifetime(&mut self, value: Duration) -> &mut Self {
         self.retry_token_lifetime = value;
         self
     }
 
-    /// Maximum number of incoming connections to buffer.
+    /// Maximum number of simultaneous connections to accept.
     ///
-    /// Accepting a connection removes it from the buffer, so this does not need to be large.
+    /// New incoming connections are only accepted if the total number of incoming or outgoing
+    /// connections is less than this. Outgoing connections are unaffected.
     pub fn concurrent_connections(&mut self, value: u32) -> &mut Self {
         self.concurrent_connections = value;
         self
@@ -511,28 +497,38 @@ where
 }
 
 #[cfg(feature = "rustls")]
-impl ServerConfig<crypto::rustls::TlsSession> {
-    /// Set the certificate chain that will be presented to clients
-    pub fn certificate(
-        &mut self,
-        cert_chain: CertificateChain,
-        key: PrivateKey,
-    ) -> Result<&mut Self, rustls::TLSError> {
-        Arc::make_mut(&mut self.crypto).set_single_cert(cert_chain.certs, key.inner)?;
-        Ok(self)
+impl ServerConfig {
+    /// Create a server config with the given certificate chain to be presented to clients
+    ///
+    /// Uses a randomized handshake token key.
+    pub fn with_single_cert(
+        cert_chain: Vec<rustls::Certificate>,
+        key: rustls::PrivateKey,
+    ) -> Result<Self, rustls::Error> {
+        let crypto = crypto::rustls::server_config(cert_chain, key)?;
+        Ok(Self::with_crypto(Arc::new(crypto)))
+    }
+
+    /// Create a server config with the given [`rustls::ServerConfig`]
+    ///
+    /// Uses a randomized handshake token key.
+    pub fn with_crypto(crypto: Arc<rustls::ServerConfig>) -> Self {
+        let rng = &mut rand::thread_rng();
+        let mut master_key = [0u8; 64];
+        rng.fill_bytes(&mut master_key);
+        let master_key = ring::hkdf::Salt::new(ring::hkdf::HKDF_SHA256, &[]).extract(&master_key);
+
+        Self::new(crypto, Arc::new(master_key))
     }
 }
 
-impl<S> fmt::Debug for ServerConfig<S>
-where
-    S: crypto::Session,
-{
+impl fmt::Debug for ServerConfig {
     fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
         fmt.debug_struct("ServerConfig<T>")
             .field("transport", &self.transport)
             .field("crypto", &"ServerConfig { elided }")
             .field("token_key", &"[ elided ]")
-            .field("use_stateless_retry", &self.use_stateless_retry)
+            .field("use_retry", &self.use_retry)
             .field("retry_token_lifetime", &self.retry_token_lifetime)
             .field("concurrent_connections", &self.concurrent_connections)
             .field("migration", &self.migration)
@@ -540,100 +536,73 @@ where
     }
 }
 
-impl<S> Default for ServerConfig<S>
-where
-    S: crypto::Session,
-{
-    fn default() -> Self {
-        let rng = &mut rand::thread_rng();
-
-        let mut master_key = [0u8; 64];
-        rng.fill_bytes(&mut master_key);
-
-        Self::new(S::HandshakeTokenKey::from_secret(&master_key))
-    }
-}
-
-impl<S> Clone for ServerConfig<S>
-where
-    S: crypto::Session,
-    S::ServerConfig: Clone,
-{
-    fn clone(&self) -> Self {
-        Self {
-            transport: self.transport.clone(),
-            crypto: self.crypto.clone(),
-            token_key: self.token_key.clone(),
-            use_stateless_retry: self.use_stateless_retry,
-            retry_token_lifetime: self.retry_token_lifetime,
-            concurrent_connections: self.concurrent_connections,
-            migration: self.migration,
-        }
-    }
-}
-
 /// Configuration for outgoing connections
 ///
 /// Default values should be suitable for most internet applications.
-pub struct ClientConfig<S>
-where
-    S: crypto::Session,
-{
+#[derive(Clone)]
+#[non_exhaustive]
+pub struct ClientConfig {
     /// Transport configuration to use
     pub transport: Arc<TransportConfig>,
 
     /// Cryptographic configuration to use
-    pub crypto: S::ClientConfig,
+    pub crypto: Arc<dyn crypto::ClientConfig>,
+
+    /// QUIC protocol version to use
+    pub(crate) version: u32,
+}
+
+impl ClientConfig {
+    /// Create a default config with a particular cryptographic config
+    pub fn new(crypto: Arc<dyn crypto::ClientConfig>) -> Self {
+        Self {
+            transport: Default::default(),
+            crypto,
+            version: 1,
+        }
+    }
+
+    /// Set the QUIC version to use
+    pub fn version(&mut self, version: u32) -> &mut Self {
+        self.version = version;
+        self
+    }
 }
 
 #[cfg(feature = "rustls")]
-impl ClientConfig<crypto::rustls::TlsSession> {
-    /// Add a trusted certificate authority
-    pub fn add_certificate_authority(
-        &mut self,
-        cert: Certificate,
-    ) -> Result<&mut Self, webpki::Error> {
-        let anchor = webpki::trust_anchor_util::cert_der_as_trust_anchor(&cert.inner.0)?;
-        Arc::make_mut(&mut self.crypto)
-            .root_store
-            .add_server_trust_anchors(&webpki::TLSServerTrustAnchors(&[anchor]));
-        Ok(self)
+impl ClientConfig {
+    /// Create a client configuration that trusts the platform's native roots
+    #[cfg(feature = "native-certs")]
+    pub fn with_native_roots() -> Self {
+        let mut roots = rustls::RootCertStore::empty();
+        match rustls_native_certs::load_native_certs() {
+            Ok(certs) => {
+                for cert in certs {
+                    if let Err(e) = roots.add(&rustls::Certificate(cert.0)) {
+                        tracing::warn!("failed to parse trust anchor: {}", e);
+                    }
+                }
+            }
+            Err(e) => {
+                tracing::warn!("couldn't load any default trust roots: {}", e);
+            }
+        };
+
+        Self::with_root_certificates(roots)
+    }
+
+    /// Create a client configuration that trusts specified trust anchors
+    pub fn with_root_certificates(roots: rustls::RootCertStore) -> Self {
+        Self::new(Arc::new(crypto::rustls::client_config(roots)))
     }
 }
 
-impl<S> Default for ClientConfig<S>
-where
-    S: crypto::Session,
-{
-    fn default() -> Self {
-        Self {
-            transport: Default::default(),
-            crypto: S::ClientConfig::new(),
-        }
-    }
-}
-
-impl<S> Clone for ClientConfig<S>
-where
-    S: crypto::Session,
-    S::ClientConfig: Clone,
-{
-    fn clone(&self) -> Self {
-        Self {
-            transport: self.transport.clone(),
-            crypto: self.crypto.clone(),
-        }
-    }
-}
-
-impl<S> fmt::Debug for ClientConfig<S>
-where
-    S: crypto::Session,
-{
+impl fmt::Debug for ClientConfig {
     fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
         fmt.debug_struct("ClientConfig<T>")
             .field("transport", &self.transport)
             .field("crypto", &"ClientConfig { elided }")
+            .field("version", &self.version)
             .finish()
     }
 }
@@ -658,3 +627,38 @@ impl From<VarIntBoundsExceeded> for ConfigError {
         ConfigError::OutOfBounds
     }
 }
+
+/// Maximum duration of inactivity to accept before timing out the connection.
+///
+/// This wraps an underlying [`VarInt`], representing the duration in milliseconds. Values can be
+/// constructed by converting directly from `VarInt`, or using `TryFrom<Duration>`.
+///
+/// ```
+/// # use std::{convert::TryFrom, time::Duration};
+/// # use quinn_proto::{IdleTimeout, VarIntBoundsExceeded, VarInt};
+/// # fn main() -> Result<(), VarIntBoundsExceeded> {
+/// // A `VarInt`-encoded value in milliseconds
+/// let timeout = IdleTimeout::from(VarInt::from_u32(10_000));
+///
+/// // Try to convert a `Duration` into a `VarInt`-encoded timeout
+/// let timeout = IdleTimeout::try_from(Duration::from_secs(10))?;
+/// # Ok(())
+/// # }
+/// ```
+#[derive(Default, Copy, Clone, Eq, Hash, Ord, PartialEq, PartialOrd)]
+pub struct IdleTimeout(VarInt);
+
+impl From<VarInt> for IdleTimeout {
+    fn from(inner: VarInt) -> Self {
+        Self(inner)
+    }
+}
+
+impl std::convert::TryFrom<Duration> for IdleTimeout {
+    type Error = VarIntBoundsExceeded;
+
+    fn try_from(timeout: Duration) -> Result<Self, Self::Error> {
+        let inner = VarInt::try_from(timeout.as_millis())?;
+        Ok(Self(inner))
+    }
+}
