@@ -1,22 +1,21 @@
 use std::{
     cmp,
-    collections::{BTreeMap, HashSet, VecDeque},
+    collections::{BTreeMap, VecDeque},
     mem,
     ops::{Index, IndexMut},
-    time::Instant,
+    time::{Duration, Instant},
 };
 
+use fxhash::FxHashSet;
+
 use super::assembler::Assembler;
 use crate::{
-    crypto, crypto::Keys, frame, packet::SpaceId, range_set::RangeSet, shared::IssuedCid, StreamId,
-    VarInt,
+    connection::StreamsState, crypto::Keys, frame, packet::SpaceId, range_set::ArrayRangeSet,
+    shared::IssuedCid, StreamId, VarInt,
 };
 
-pub(crate) struct PacketSpace<S>
-where
-    S: crypto::Session,
-{
-    pub(crate) crypto: Option<Keys<S>>,
+pub(crate) struct PacketSpace {
+    pub(crate) crypto: Option<Keys>,
     pub(crate) dedup: Dedup,
     /// Highest received packet number
     pub(crate) rx_packet: u64,
@@ -24,9 +23,7 @@ where
     /// Data to send
     pub(crate) pending: Retransmits,
     /// Packet numbers to acknowledge
-    pub(crate) pending_acks: RangeSet,
-    /// Set iff we have received a non-ack frame since the last ack-only packet we sent
-    pub(crate) permit_ack_only: bool,
+    pub(crate) pending_acks: PendingAcks,
 
     /// The packet number of the next packet that will be sent, if any.
     pub(crate) next_packet_number: u64,
@@ -66,10 +63,7 @@ where
     pub(crate) sent_with_keys: u64,
 }
 
-impl<S> PacketSpace<S>
-where
-    S: crypto::Session,
-{
+impl PacketSpace {
     pub(crate) fn new(now: Instant) -> Self {
         Self {
             crypto: None,
@@ -77,8 +71,7 @@ where
             rx_packet: 0,
 
             pending: Retransmits::default(),
-            pending_acks: RangeSet::new(),
-            permit_ack_only: false,
+            pending_acks: PendingAcks::default(),
 
             next_packet_number: 0,
             largest_acked_packet: None,
@@ -111,19 +104,19 @@ where
     /// waiting to be sent, then we retransmit in-flight data to reduce odds of loss. If there's no
     /// in-flight data either, we're probably a client guarding against a handshake
     /// anti-amplification deadlock and we just make something up.
-    pub(crate) fn maybe_queue_probe(&mut self) {
+    pub(crate) fn maybe_queue_probe(&mut self, streams: &StreamsState) {
         if self.loss_probes == 0 {
             return;
         }
 
         // Retransmit the data of the oldest in-flight packet
-        if !self.pending.is_empty() {
+        if !self.pending.is_empty(streams) {
             // There's real data to send here, no need to make something up
             return;
         }
 
         for packet in self.sent_packets.values_mut() {
-            if !packet.retransmits.is_empty() {
+            if !packet.retransmits.is_empty(streams) {
                 // Remove retransmitted data from the old packet so we don't end up retransmitting
                 // it *again* even if the copy we're sending now gets acknowledged.
                 self.pending |= mem::take(&mut packet.retransmits);
@@ -146,10 +139,11 @@ where
         x
     }
 
-    pub(crate) fn can_send(&self) -> bool {
-        !self.pending.is_empty()
-            || (self.permit_ack_only && !self.pending_acks.is_empty())
-            || self.ping_pending
+    pub(crate) fn can_send(&self, streams: &StreamsState) -> SendableFrames {
+        let acks = self.pending_acks.can_send();
+        let other = !self.pending.is_empty(streams) || self.ping_pending;
+
+        SendableFrames { acks, other }
     }
 
     /// Verifies sanity of an ECN block and returns whether congestion was encountered.
@@ -191,15 +185,15 @@ where
     }
 }
 
-impl<S: crypto::Session> Index<SpaceId> for [PacketSpace<S>; 3] {
-    type Output = PacketSpace<S>;
-    fn index(&self, space: SpaceId) -> &PacketSpace<S> {
+impl Index<SpaceId> for [PacketSpace; 3] {
+    type Output = PacketSpace;
+    fn index(&self, space: SpaceId) -> &PacketSpace {
         &self.as_ref()[space as usize]
     }
 }
 
-impl<S: crypto::Session> IndexMut<SpaceId> for [PacketSpace<S>; 3] {
-    fn index_mut(&mut self, space: SpaceId) -> &mut PacketSpace<S> {
+impl IndexMut<SpaceId> for [PacketSpace; 3] {
+    fn index_mut(&mut self, space: SpaceId) -> &mut PacketSpace {
         &mut self.as_mut()[space as usize]
     }
 }
@@ -215,7 +209,7 @@ pub(crate) struct SentPacket {
     pub(crate) size: u16,
     /// Whether an acknowledgement is expected directly in response to this packet.
     pub(crate) ack_eliciting: bool,
-    pub(crate) acks: RangeSet,
+    pub(crate) acks: ArrayRangeSet,
     /// Data which needs to be retransmitted in case the packet is lost.
     /// The data is boxed to minimize `SentPacket` size for the typical case of
     /// packets only containing ACKs and STREAM frames.
@@ -234,7 +228,7 @@ pub struct Retransmits {
     pub(crate) max_bi_stream_id: bool,
     pub(crate) reset_stream: Vec<(StreamId, VarInt)>,
     pub(crate) stop_sending: Vec<frame::StopSending>,
-    pub(crate) max_stream_data: HashSet<StreamId>,
+    pub(crate) max_stream_data: FxHashSet<StreamId>,
     pub(crate) crypto: VecDeque<frame::Crypto>,
     pub(crate) new_cids: Vec<IssuedCid>,
     pub(crate) retire_cids: Vec<u64>,
@@ -242,13 +236,16 @@ pub struct Retransmits {
 }
 
 impl Retransmits {
-    pub fn is_empty(&self) -> bool {
+    pub fn is_empty(&self, streams: &StreamsState) -> bool {
         !self.max_data
             && !self.max_uni_stream_id
             && !self.max_bi_stream_id
             && self.reset_stream.is_empty()
             && self.stop_sending.is_empty()
-            && self.max_stream_data.is_empty()
+            && self
+                .max_stream_data
+                .iter()
+                .all(|&id| !streams.can_send_flow_control(id))
             && self.crypto.is_empty()
             && self.new_cids.is_empty()
             && self.retire_cids.is_empty()
@@ -264,7 +261,7 @@ impl Default for Retransmits {
             max_bi_stream_id: false,
             reset_stream: Vec::new(),
             stop_sending: Vec::new(),
-            max_stream_data: HashSet::new(),
+            max_stream_data: FxHashSet::default(),
             crypto: VecDeque::new(),
             new_cids: Vec::new(),
             retire_cids: Vec::new(),
@@ -321,9 +318,9 @@ pub struct ThinRetransmits {
 
 impl ThinRetransmits {
     /// Returns `true` if no retransmits are necessary
-    pub fn is_empty(&self) -> bool {
+    pub fn is_empty(&self, streams: &StreamsState) -> bool {
         match &self.retransmits {
-            Some(retransmits) => retransmits.is_empty(),
+            Some(retransmits) => retransmits.is_empty(streams),
             None => true,
         }
     }
@@ -411,6 +408,98 @@ impl Dedup {
     }
 }
 
+/// Indicates which data is available for sending
+#[derive(Clone, Copy, PartialEq, Eq, Debug)]
+pub(crate) struct SendableFrames {
+    pub acks: bool,
+    pub other: bool,
+}
+
+impl SendableFrames {
+    /// Returns that no data is available for sending
+    pub fn empty() -> Self {
+        Self {
+            acks: false,
+            other: false,
+        }
+    }
+
+    /// Whether no data is sendable
+    pub fn is_empty(&self) -> bool {
+        !self.acks && !self.other
+    }
+}
+
+#[derive(Debug, Default)]
+pub(crate) struct PendingAcks {
+    permit_ack_only: bool,
+    ranges: ArrayRangeSet,
+    /// This value will be used for calculating ACK delay once it is implemented
+    ///
+    /// ACK delay will be the delay between when a packet arrived (`latest_incoming`)
+    /// and between it will be allowed to be acknowledged (`can_send() == true`).
+    latest_incoming: Option<Instant>,
+    ack_delay: Duration,
+}
+
+impl PendingAcks {
+    /// Whether any ACK frames can be sent
+    pub fn can_send(&self) -> bool {
+        self.permit_ack_only && !self.ranges.is_empty()
+    }
+
+    /// Returns the duration the acknowledgement of the latest incoming packet has been delayed
+    pub fn ack_delay(&self) -> Duration {
+        self.ack_delay
+    }
+
+    /// Should be called whenever an ACK eliciting frame was received
+    ///
+    /// This requires sending new outgoing ACKs
+    pub fn ack_eliciting_frame_received(&mut self) {
+        self.permit_ack_only = true;
+    }
+
+    /// Should be called whenever ACKs have been sent
+    ///
+    /// This will suppress sending further ACKs until additional ACK eliciting frames arrive
+    pub fn acks_sent(&mut self) {
+        // If we sent any acks, don't immediately resend them. Setting this even if ack_only is
+        // false needlessly prevents us from ACKing the next packet if it's ACK-only, but saves
+        // the need for subtler logic to avoid double-transmitting acks all the time.
+        // This reset needs to happen before we check whether more data
+        // is available in this space - because otherwise it would return
+        // `true` purely due to the ACKs
+        self.permit_ack_only = false;
+    }
+
+    /// Insert one packet that needs to be acknowledged
+    pub fn insert_one(&mut self, packet: u64, now: Instant) {
+        self.ranges.insert_one(packet);
+        self.latest_incoming = Some(now);
+
+        if self.ranges.len() > MAX_ACK_BLOCKS {
+            self.ranges.pop_min();
+        }
+    }
+
+    /// Removes the given ACKs from the set of pending ACKs
+    pub fn subtract(&mut self, acks: &ArrayRangeSet) {
+        self.ranges.subtract(acks);
+        if self.ranges.is_empty() {
+            self.permit_ack_only = false;
+        }
+    }
+
+    /// Returns the set of currently pending ACK ranges
+    pub fn ranges(&self) -> &ArrayRangeSet {
+        &self.ranges
+    }
+}
+
+/// Ensures we can always fit all our ACKs in a single minimum-MTU packet with room to spare
+const MAX_ACK_BLOCKS: usize = 64;
+
 #[cfg(test)]
 mod test {
     use super::*;
