@@ -5,10 +5,11 @@ use std::{
     net::{IpAddr, SocketAddr},
     ops::{Index, IndexMut},
     sync::Arc,
-    time::{Duration, Instant, SystemTime},
+    time::{Instant, SystemTime},
 };
 
 use bytes::{BufMut, Bytes, BytesMut};
+use fxhash::FxHashMap;
 use rand::{rngs::StdRng, Rng, RngCore, SeedableRng};
 use slab::Slab;
 use thiserror::Error;
@@ -17,12 +18,9 @@ use tracing::{debug, trace, warn};
 use crate::{
     cid_generator::{ConnectionIdGenerator, RandomConnectionIdGenerator},
     coding::BufMutExt,
-    config::{ClientConfig, ConfigError, EndpointConfig, ServerConfig},
+    config::{ClientConfig, EndpointConfig, ServerConfig},
     connection::{Connection, ConnectionError},
-    crypto::{
-        self, ClientConfig as ClientCryptoConfig, Keys, PacketKey,
-        ServerConfig as ServerCryptoConfig,
-    },
+    crypto::{self, Keys, UnsupportedVersion},
     frame,
     packet::{Header, Packet, PacketDecodeError, PacketNumber, PartialDecode},
     shared::{
@@ -30,8 +28,8 @@ use crate::{
         EndpointEventInner, IssuedCid,
     },
     transport_parameters::TransportParameters,
-    ResetToken, RetryToken, Side, Transmit, TransportError, MAX_CID_SIZE, MIN_INITIAL_SIZE,
-    MIN_MTU, RESET_TOKEN_SIZE,
+    ResetToken, RetryToken, Side, Transmit, TransportConfig, TransportError,
+    INITIAL_MAX_UDP_PAYLOAD_SIZE, MAX_CID_SIZE, MIN_INITIAL_SIZE, RESET_TOKEN_SIZE,
 };
 
 /// The main entry point to the library
@@ -39,15 +37,20 @@ use crate::{
 /// This object performs no I/O whatsoever. Instead, it generates a stream of packets to send via
 /// `poll_transmit`, and consumes incoming packets and connection-generated events via `handle` and
 /// `handle_event`.
-pub struct Endpoint<S>
-where
-    S: crypto::Session,
-{
+pub struct Endpoint {
     rng: StdRng,
     transmits: VecDeque<Transmit>,
+    /// Identifies connections based on the initial DCID the peer utilized
+    ///
+    /// Uses a standard `HashMap` to protect against hash collision attacks.
     connection_ids_initial: HashMap<ConnectionId, ConnectionHandle>,
-    connection_ids: HashMap<ConnectionId, ConnectionHandle>,
+    /// Identifies connections based on locally created CIDs
+    ///
+    /// Uses a cheaper hash function since keys are locally created
+    connection_ids: FxHashMap<ConnectionId, ConnectionHandle>,
     /// Identifies connections with zero-length CIDs
+    ///
+    /// Uses a standard `HashMap` to protect against hash collision attacks.
     connection_remotes: HashMap<SocketAddr, ConnectionHandle>,
     /// Reset tokens provided by the peer for the CID each connection is currently sending to
     ///
@@ -56,31 +59,25 @@ where
     connection_reset_tokens: ResetTokenTable,
     connections: Slab<ConnectionMeta>,
     local_cid_generator: Box<dyn ConnectionIdGenerator>,
-    config: Arc<EndpointConfig<S>>,
-    server_config: Option<Arc<ServerConfig<S>>>,
+    config: Arc<EndpointConfig>,
+    server_config: Option<Arc<ServerConfig>>,
     /// Whether incoming connections should be unconditionally rejected by a server
     ///
     /// Equivalent to a `ServerConfig.accept_buffer` of `0`, but can be changed after the endpoint is constructed.
     reject_new_connections: bool,
 }
 
-impl<S> Endpoint<S>
-where
-    S: crypto::Session,
-{
+impl Endpoint {
     /// Create a new endpoint
     ///
     /// Returns `Err` if the configuration is invalid.
-    pub fn new(
-        config: Arc<EndpointConfig<S>>,
-        server_config: Option<Arc<ServerConfig<S>>>,
-    ) -> Self {
+    pub fn new(config: Arc<EndpointConfig>, server_config: Option<Arc<ServerConfig>>) -> Self {
         Self {
             rng: StdRng::from_entropy(),
             transmits: VecDeque::new(),
-            connection_ids_initial: HashMap::new(),
-            connection_ids: HashMap::new(),
-            connection_remotes: HashMap::new(),
+            connection_ids_initial: HashMap::default(),
+            connection_ids: FxHashMap::default(),
+            connection_remotes: HashMap::default(),
             connection_reset_tokens: ResetTokenTable::default(),
             connections: Slab::new(),
             local_cid_generator: (config.connection_id_generator_factory.as_ref())(),
@@ -90,16 +87,17 @@ where
         }
     }
 
-    fn is_server(&self) -> bool {
-        self.server_config.is_some()
-    }
-
     /// Get the next packet to transmit
     #[must_use]
     pub fn poll_transmit(&mut self) -> Option<Transmit> {
         self.transmits.pop_front()
     }
 
+    /// Replace the server configuration, affecting new incoming connections only
+    pub fn set_server_config(&mut self, server_config: Option<Arc<ServerConfig>>) {
+        self.server_config = server_config;
+    }
+
     /// Process `EndpointEvent`s emitted from related `Connection`s
     ///
     /// In turn, processing this event may return a `ConnectionEvent` for the same `Connection`.
@@ -136,7 +134,7 @@ where
                     self.connection_ids_initial.remove(&conn.init_cid);
                 }
                 for cid in conn.loc_cids.values() {
-                    self.connection_ids.remove(&cid);
+                    self.connection_ids.remove(cid);
                 }
                 self.connection_remotes.remove(&conn.initial_remote);
                 if let Some((remote, token)) = conn.reset_token {
@@ -155,7 +153,7 @@ where
         local_ip: Option<IpAddr>,
         ecn: Option<EcnCodepoint>,
         data: BytesMut,
-    ) -> Option<(ConnectionHandle, DatagramEvent<S>)> {
+    ) -> Option<(ConnectionHandle, DatagramEvent)> {
         let datagram_len = data.len();
         let (first_decode, remaining) = match PartialDecode::new(
             data,
@@ -168,7 +166,7 @@ where
                 dst_cid,
                 version,
             }) => {
-                if !self.is_server() {
+                if self.server_config.is_none() {
                     debug!("dropping packet with unsupported version");
                     return None;
                 }
@@ -187,7 +185,9 @@ where
                 } else {
                     buf.write::<u32>(0x0a1a_2a4a);
                 }
-                buf.write(self.config.initial_version); // supported version
+                for &version in &self.config.supported_versions {
+                    buf.write(version);
+                }
                 self.transmits.push_back(Transmit {
                     destination: remote,
                     ecn: None,
@@ -255,27 +255,37 @@ where
         // Potentially create a new connection
         //
 
-        if !self.is_server() {
-            debug!("packet for unrecognized connection {}", dst_cid);
-            self.stateless_reset(datagram_len, remote, local_ip, &dst_cid);
-            return None;
-        }
-
-        if first_decode.has_long_header() {
-            if !first_decode.is_initial() {
-                debug!(
-                    "ignoring non-initial packet for unknown connection {}",
-                    dst_cid
-                );
+        let server_config = match &self.server_config {
+            Some(config) => config,
+            None => {
+                debug!("packet for unrecognized connection {}", dst_cid);
+                self.stateless_reset(datagram_len, remote, local_ip, &dst_cid);
                 return None;
             }
+        };
+
+        if let Some(version) = first_decode.initial_version() {
             if datagram_len < MIN_INITIAL_SIZE as usize {
                 debug!("ignoring short initial for connection {}", dst_cid);
                 return None;
             }
 
-            let crypto = S::initial_keys(&dst_cid, Side::Server);
-            return match first_decode.finish(Some(&crypto.header.remote)) {
+            let crypto = match server_config
+                .crypto
+                .initial_keys(version, &dst_cid, Side::Server)
+            {
+                Ok(keys) => keys,
+                Err(UnsupportedVersion) => {
+                    // This probably indicates that the user set supported_versions incorrectly in
+                    // `EndpointConfig`.
+                    debug!(
+                        "ignoring initial packet version {:#x} unsupported by cryptographic layer",
+                        version
+                    );
+                    return None;
+                }
+            };
+            return match first_decode.finish(Some(&*crypto.header.remote)) {
                 Ok(packet) => self
                     .handle_first_packet(now, remote, local_ip, ecn, packet, remaining, &crypto)
                     .map(|(ch, conn)| (ch, DatagramEvent::NewConnection(conn))),
@@ -284,6 +294,12 @@ where
                     None
                 }
             };
+        } else if first_decode.has_long_header() {
+            debug!(
+                "ignoring non-initial packet for unknown connection {}",
+                dst_cid
+            );
+            return None;
         }
 
         //
@@ -348,29 +364,44 @@ where
     /// Initiate a connection
     pub fn connect(
         &mut self,
-        config: ClientConfig<S>,
+        config: ClientConfig,
         remote: SocketAddr,
         server_name: &str,
-    ) -> Result<(ConnectionHandle, Connection<S>), ConnectError> {
+    ) -> Result<(ConnectionHandle, Connection), ConnectError> {
         if self.is_full() {
             return Err(ConnectError::TooManyConnections);
         }
         if remote.port() == 0 {
             return Err(ConnectError::InvalidRemoteAddress(remote));
         }
+
         let remote_id = RandomConnectionIdGenerator::new(MAX_CID_SIZE).generate_cid();
         trace!(initial_dcid = %remote_id);
+
+        let loc_cid = self.new_cid();
+        let params = TransportParameters::new(
+            &config.transport,
+            &self.config,
+            self.local_cid_generator.as_ref(),
+            loc_cid,
+            None,
+        );
+        let tls = config
+            .crypto
+            .start_session(config.version, server_name, &params)?;
+
         let (ch, conn) = self.add_connection(
+            config.version,
             remote_id,
+            loc_cid,
             remote_id,
             remote,
             None,
-            ConnectionOpts::Client {
-                config,
-                server_name: server_name.into(),
-            },
             Instant::now(),
-        )?;
+            tls,
+            None,
+            config.transport,
+        );
         Ok((ch, conn))
     }
 
@@ -407,90 +438,6 @@ where
         }
     }
 
-    fn add_connection(
-        &mut self,
-        init_cid: ConnectionId,
-        rem_cid: ConnectionId,
-        remote: SocketAddr,
-        local_ip: Option<IpAddr>,
-        opts: ConnectionOpts<S>,
-        now: Instant,
-    ) -> Result<(ConnectionHandle, Connection<S>), ConnectError> {
-        let loc_cid = self.new_cid();
-        let (server_config, tls, transport_config) = match opts {
-            ConnectionOpts::Client {
-                config,
-                server_name,
-            } => {
-                let params = TransportParameters::new::<S>(
-                    &config.transport,
-                    &self.config,
-                    self.local_cid_generator.as_ref(),
-                    loc_cid,
-                    None,
-                );
-                (
-                    None,
-                    config.crypto.start_session(&server_name, &params)?,
-                    config.transport,
-                )
-            }
-            ConnectionOpts::Server {
-                orig_dst_cid,
-                retry_src_cid,
-            } => {
-                let config = self.server_config.as_ref().unwrap();
-                let params = TransportParameters::new(
-                    &config.transport,
-                    &self.config,
-                    self.local_cid_generator.as_ref(),
-                    loc_cid,
-                    Some(config),
-                );
-                let server_params = TransportParameters {
-                    stateless_reset_token: Some(ResetToken::new(&*self.config.reset_key, &loc_cid)),
-                    original_dst_cid: Some(orig_dst_cid),
-                    retry_src_cid,
-                    ..params
-                };
-                (
-                    Some(config.clone()),
-                    config.crypto.start_session(&server_params),
-                    config.transport.clone(),
-                )
-            }
-        };
-
-        let conn = Connection::new(
-            server_config,
-            transport_config,
-            init_cid,
-            loc_cid,
-            rem_cid,
-            remote,
-            local_ip,
-            tls,
-            self.local_cid_generator.as_ref(),
-            now,
-            self.config.initial_version,
-        );
-        let id = self.connections.insert(ConnectionMeta {
-            init_cid,
-            cids_issued: 0,
-            loc_cids: iter::once((0, loc_cid)).collect(),
-            initial_remote: remote,
-            reset_token: None,
-        });
-        let ch = ConnectionHandle(id);
-
-        if self.local_cid_generator.cid_len() > 0 {
-            self.connection_ids.insert(loc_cid, ch);
-        } else {
-            self.connection_remotes.insert(remote, ch);
-        }
-        Ok((ch, conn))
-    }
-
     fn handle_first_packet(
         &mut self,
         now: Instant,
@@ -499,16 +446,17 @@ where
         ecn: Option<EcnCodepoint>,
         mut packet: Packet,
         rest: Option<BytesMut>,
-        crypto: &Keys<S>,
-    ) -> Option<(ConnectionHandle, Connection<S>)> {
-        let (src_cid, dst_cid, token, packet_number) = match packet.header {
+        crypto: &Keys,
+    ) -> Option<(ConnectionHandle, Connection)> {
+        let (src_cid, dst_cid, token, packet_number, version) = match packet.header {
             Header::Initial {
                 src_cid,
                 dst_cid,
                 ref token,
                 number,
+                version,
                 ..
-            } => (src_cid, dst_cid, token.clone(), number),
+            } => (src_cid, dst_cid, token.clone(), number, version),
             _ => panic!("non-initial packet in handle_first_packet()"),
         };
         let packet_number = packet_number.expand(0);
@@ -532,8 +480,7 @@ where
             return None;
         }
 
-        // Local CID used for stateless packets
-        let temp_loc_cid = self.new_cid();
+        let loc_cid = self.new_cid();
         let server_config = self.server_config.as_ref().unwrap();
 
         if self.connections.len() >= server_config.concurrent_connections as usize
@@ -542,36 +489,37 @@ where
         {
             debug!("refusing connection");
             self.initial_close(
+                version,
                 remote,
                 local_ip,
                 crypto,
                 &src_cid,
-                &temp_loc_cid,
+                &loc_cid,
                 TransportError::CONNECTION_REFUSED(""),
             );
             return None;
         }
 
         if dst_cid.len() < 8
-            && (!server_config.use_stateless_retry
-                || dst_cid.len() != self.local_cid_generator.cid_len())
+            && (!server_config.use_retry || dst_cid.len() != self.local_cid_generator.cid_len())
         {
             debug!(
                 "rejecting connection due to invalid DCID length {}",
                 dst_cid.len()
             );
             self.initial_close(
+                version,
                 remote,
                 local_ip,
                 crypto,
                 &src_cid,
-                &temp_loc_cid,
+                &loc_cid,
                 TransportError::PROTOCOL_VIOLATION("invalid destination CID length"),
             );
             return None;
         }
 
-        let (retry_src_cid, orig_dst_cid) = if server_config.use_stateless_retry {
+        let (retry_src_cid, orig_dst_cid) = if server_config.use_retry {
             if token.is_empty() {
                 // First Initial
                 let mut random_bytes = vec![0u8; RetryToken::RANDOM_BYTES_LEN];
@@ -582,19 +530,19 @@ where
                     issued: SystemTime::now(),
                     random_bytes: &random_bytes,
                 }
-                .encode(&*server_config.token_key, &remote, &temp_loc_cid);
+                .encode(&*server_config.token_key, &remote, &loc_cid);
 
                 let header = Header::Retry {
-                    src_cid: temp_loc_cid,
+                    src_cid: loc_cid,
                     dst_cid: src_cid,
-                    version: self.config.initial_version,
+                    version,
                 };
 
                 let mut buf = Vec::new();
                 let encode = header.encode(&mut buf);
                 buf.put_slice(&token);
-                buf.extend_from_slice(&S::retry_tag(&dst_cid, &buf));
-                encode.finish::<S::PacketKey, S::HeaderKey>(&mut buf, &crypto.header.local, None);
+                buf.extend_from_slice(&server_config.crypto.retry_tag(version, &dst_cid, &buf));
+                encode.finish(&mut buf, &*crypto.header.local, None);
 
                 self.transmits.push_back(Transmit {
                     destination: remote,
@@ -608,22 +556,19 @@ where
 
             match RetryToken::from_bytes(&*server_config.token_key, &remote, &dst_cid, &token) {
                 Ok(token)
-                    if token.issued
-                        + Duration::from_micros(
-                            self.server_config.as_ref().unwrap().retry_token_lifetime,
-                        )
-                        > SystemTime::now() =>
+                    if token.issued + server_config.retry_token_lifetime > SystemTime::now() =>
                 {
                     (Some(dst_cid), token.orig_dst_cid)
                 }
                 _ => {
                     debug!("rejecting invalid stateless retry token");
                     self.initial_close(
+                        version,
                         remote,
                         local_ip,
                         crypto,
                         &src_cid,
-                        &temp_loc_cid,
+                        &loc_cid,
                         TransportError::INVALID_TOKEN(""),
                     );
                     return None;
@@ -633,19 +578,32 @@ where
             (None, dst_cid)
         };
 
-        let (ch, mut conn) = self
-            .add_connection(
-                dst_cid,
-                src_cid,
-                remote,
-                local_ip,
-                ConnectionOpts::Server {
-                    retry_src_cid,
-                    orig_dst_cid,
-                },
-                now,
-            )
-            .unwrap();
+        let server_config = server_config.clone();
+        let mut params = TransportParameters::new(
+            &server_config.transport,
+            &self.config,
+            self.local_cid_generator.as_ref(),
+            loc_cid,
+            Some(&server_config),
+        );
+        params.stateless_reset_token = Some(ResetToken::new(&*self.config.reset_key, &loc_cid));
+        params.original_dst_cid = Some(orig_dst_cid);
+        params.retry_src_cid = retry_src_cid;
+
+        let tls = server_config.crypto.clone().start_session(version, &params);
+        let transport_config = server_config.transport.clone();
+        let (ch, mut conn) = self.add_connection(
+            version,
+            dst_cid,
+            loc_cid,
+            src_cid,
+            remote,
+            local_ip,
+            now,
+            tls,
+            Some(server_config),
+            transport_config,
+        );
         if dst_cid.len() != 0 {
             self.connection_ids_initial.insert(dst_cid, ch);
         }
@@ -658,18 +616,63 @@ where
                 debug!("handshake failed: {}", e);
                 self.handle_event(ch, EndpointEvent(EndpointEventInner::Drained));
                 if let ConnectionError::TransportError(e) = e {
-                    self.initial_close(remote, local_ip, crypto, &src_cid, &temp_loc_cid, e);
+                    self.initial_close(version, remote, local_ip, crypto, &src_cid, &loc_cid, e);
                 }
                 None
             }
         }
     }
 
+    fn add_connection(
+        &mut self,
+        version: u32,
+        init_cid: ConnectionId,
+        loc_cid: ConnectionId,
+        rem_cid: ConnectionId,
+        remote: SocketAddr,
+        local_ip: Option<IpAddr>,
+        now: Instant,
+        tls: Box<dyn crypto::Session>,
+        server_config: Option<Arc<ServerConfig>>,
+        transport_config: Arc<TransportConfig>,
+    ) -> (ConnectionHandle, Connection) {
+        let conn = Connection::new(
+            server_config,
+            transport_config,
+            init_cid,
+            loc_cid,
+            rem_cid,
+            remote,
+            local_ip,
+            tls,
+            self.local_cid_generator.as_ref(),
+            now,
+            version,
+        );
+
+        let id = self.connections.insert(ConnectionMeta {
+            init_cid,
+            cids_issued: 0,
+            loc_cids: iter::once((0, loc_cid)).collect(),
+            initial_remote: remote,
+            reset_token: None,
+        });
+
+        let ch = ConnectionHandle(id);
+        match self.local_cid_generator.cid_len() {
+            0 => self.connection_remotes.insert(remote, ch),
+            _ => self.connection_ids.insert(loc_cid, ch),
+        };
+
+        (ch, conn)
+    }
+
     fn initial_close(
         &mut self,
+        version: u32,
         destination: SocketAddr,
         local_ip: Option<IpAddr>,
-        crypto: &Keys<S>,
+        crypto: &Keys,
         remote_id: &ConnectionId,
         local_id: &ConnectionId,
         reason: TransportError,
@@ -680,18 +683,20 @@ where
             src_cid: *local_id,
             number,
             token: Bytes::new(),
-            version: self.config.initial_version,
+            version,
         };
 
         let mut buf = Vec::<u8>::new();
         let partial_encode = header.encode(&mut buf);
-        let max_len = MIN_MTU as usize - partial_encode.header_len - crypto.packet.local.tag_len();
+        let max_len = INITIAL_MAX_UDP_PAYLOAD_SIZE as usize
+            - partial_encode.header_len
+            - crypto.packet.local.tag_len();
         frame::Close::from(reason).encode(&mut buf, max_len);
         buf.resize(buf.len() + crypto.packet.local.tag_len(), 0);
         partial_encode.finish(
             &mut buf,
-            &crypto.header.local,
-            Some((0, &crypto.packet.local)),
+            &*crypto.header.local,
+            Some((0, &*crypto.packet.local)),
         );
         self.transmits.push_back(Transmit {
             destination,
@@ -708,7 +713,7 @@ where
     }
 
     /// Access the configuration used by this endpoint
-    pub fn config(&self) -> &EndpointConfig<S> {
+    pub fn config(&self) -> &EndpointConfig {
         &self.config
     }
 
@@ -741,10 +746,7 @@ where
     }
 }
 
-impl<S> fmt::Debug for Endpoint<S>
-where
-    S: crypto::Session,
-{
+impl fmt::Debug for Endpoint {
     fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
         fmt.debug_struct("Endpoint<T>")
             .field("rng", &self.rng)
@@ -766,7 +768,7 @@ pub(crate) struct ConnectionMeta {
     init_cid: ConnectionId,
     /// Number of local connection IDs that have been issued in NEW_CONNECTION_ID frames.
     cids_issued: u64,
-    loc_cids: HashMap<u64, ConnectionId>,
+    loc_cids: FxHashMap<u64, ConnectionId>,
     /// Remote address the connection began with
     ///
     /// Only needed to support connections with zero-length CIDs, which cannot migrate, so we don't
@@ -801,25 +803,12 @@ impl IndexMut<ConnectionHandle> for Slab<ConnectionMeta> {
 }
 
 /// Event resulting from processing a single datagram
-pub enum DatagramEvent<S>
-where
-    S: crypto::Session,
-{
+#[allow(clippy::large_enum_variant)] // Not passed around extensively
+pub enum DatagramEvent {
     /// The datagram is redirected to its `Connection`
     ConnectionEvent(ConnectionEvent),
     /// The datagram has resulted in starting a new `Connection`
-    NewConnection(Connection<S>),
-}
-
-enum ConnectionOpts<S: crypto::Session> {
-    Client {
-        config: ClientConfig<S>,
-        server_name: String,
-    },
-    Server {
-        retry_src_cid: Option<ConnectionId>,
-        orig_dst_cid: ConnectionId,
-    },
+    NewConnection(Connection),
 }
 
 /// Errors in the parameters being used to create a new connection
@@ -840,16 +829,25 @@ pub enum ConnectError {
     /// The domain name supplied was malformed
     #[error("invalid DNS name: {0}")]
     InvalidDnsName(String),
-    /// The transport configuration was invalid
-    #[error("transport configuration error: {0}")]
-    Config(#[source] ConfigError),
     /// The remote [`SocketAddr`] supplied was malformed
     ///
     /// Examples include attempting to connect to port 0, or using an inappropriate address family.
     #[error("invalid remote address: {0}")]
     InvalidRemoteAddress(SocketAddr),
+    /// No default client configuration was set up
+    ///
+    /// Use `Endpoint::connect_with` to specify a client configuration.
+    #[error("no default client config")]
+    NoDefaultClientConfig,
+    /// The cryptographic layer does not support the specified QUIC version
+    #[error("unsupported QUIC version")]
+    UnsupportedVersion,
 }
 
+/// Reset Tokens which are associated with peer socket addresses
+///
+/// The standard `HashMap` is used since both `SocketAddr` and `ResetToken` are
+/// peer generated and might be usable for hash collision attacks.
 #[derive(Default, Debug)]
 struct ResetTokenTable(HashMap<SocketAddr, HashMap<ResetToken, ConnectionHandle>>);
 
