@@ -1,5 +1,6 @@
 use std::{
-    fmt, io, mem,
+    fmt::{self, Write},
+    io, mem,
     ops::{Range, RangeInclusive},
 };
 
@@ -8,7 +9,7 @@ use tinyvec::TinyVec;
 
 use crate::{
     coding::{self, BufExt, BufMutExt, UnexpectedEnd},
-    range_set::RangeSet,
+    range_set::ArrayRangeSet,
     shared::{ConnectionId, EcnCodepoint},
     Dir, ResetToken, StreamId, TransportError, TransportErrorCode, VarInt, MAX_CID_SIZE,
     RESET_TOKEN_SIZE,
@@ -324,7 +325,7 @@ impl ApplicationClose {
     }
 }
 
-#[derive(Debug, Clone, Eq, PartialEq)]
+#[derive(Clone, Eq, PartialEq)]
 pub struct Ack {
     pub largest: u64,
     pub delay: u64,
@@ -332,6 +333,28 @@ pub struct Ack {
     pub ecn: Option<EcnCounts>,
 }
 
+impl fmt::Debug for Ack {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let mut ranges = "[".to_string();
+        let mut first = true;
+        for range in self.iter() {
+            if !first {
+                ranges.push(',');
+            }
+            write!(ranges, "{:?}", range).unwrap();
+            first = false;
+        }
+        ranges.push(']');
+
+        f.debug_struct("Ack")
+            .field("largest", &self.largest)
+            .field("delay", &self.delay)
+            .field("ecn", &self.ecn)
+            .field("ranges", &ranges)
+            .finish()
+    }
+}
+
 impl<'a> IntoIterator for &'a Ack {
     type Item = RangeInclusive<u64>;
     type IntoIter = AckIter<'a>;
@@ -342,7 +365,12 @@ impl<'a> IntoIterator for &'a Ack {
 }
 
 impl Ack {
-    pub fn encode<W: BufMut>(delay: u64, ranges: &RangeSet, ecn: Option<&EcnCounts>, buf: &mut W) {
+    pub fn encode<W: BufMut>(
+        delay: u64,
+        ranges: &ArrayRangeSet,
+        ecn: Option<&EcnCounts>,
+        buf: &mut W,
+    ) {
         let mut rest = ranges.iter().rev();
         let first = rest.next().unwrap();
         let largest = first.end - 1;
@@ -383,13 +411,13 @@ pub struct EcnCounts {
 impl std::ops::AddAssign<EcnCodepoint> for EcnCounts {
     fn add_assign(&mut self, rhs: EcnCodepoint) {
         match rhs {
-            EcnCodepoint::ECT0 => {
+            EcnCodepoint::Ect0 => {
                 self.ect0 += 1;
             }
-            EcnCodepoint::ECT1 => {
+            EcnCodepoint::Ect1 => {
                 self.ect1 += 1;
             }
-            EcnCodepoint::CE => {
+            EcnCodepoint::Ce => {
                 self.ce += 1;
             }
         }
@@ -849,7 +877,7 @@ mod test {
     #[allow(clippy::range_plus_one)]
     fn ack_coding() {
         const PACKETS: &[u64] = &[1, 2, 3, 5, 10, 11, 14];
-        let mut ranges = RangeSet::new();
+        let mut ranges = ArrayRangeSet::new();
         for &packet in PACKETS {
             ranges.insert(packet..packet + 1);
         }
