@@ -52,7 +52,14 @@ impl PartialDecode {
 
     /// The underlying partially-decoded packet data
     pub(crate) fn data(&self) -> &[u8] {
-        &self.buf.get_ref()
+        self.buf.get_ref()
+    }
+
+    pub(crate) fn initial_version(&self) -> Option<u32> {
+        match self.plain_header {
+            PlainHeader::Initial { version, .. } => Some(version),
+            _ => None,
+        }
     }
 
     pub(crate) fn has_long_header(&self) -> bool {
@@ -96,10 +103,10 @@ impl PartialDecode {
         self.buf.get_ref().len()
     }
 
-    pub(crate) fn finish<H>(self, header_crypto: Option<&H>) -> Result<Packet, PacketDecodeError>
-    where
-        H: crypto::HeaderKey,
-    {
+    pub(crate) fn finish(
+        self,
+        header_crypto: Option<&dyn crypto::HeaderKey>,
+    ) -> Result<Packet, PacketDecodeError> {
         use self::PlainHeader::*;
         let Self {
             plain_header,
@@ -187,13 +194,10 @@ impl PartialDecode {
         })
     }
 
-    fn decrypt_header<H>(
+    fn decrypt_header(
         buf: &mut io::Cursor<BytesMut>,
-        header_crypto: &H,
-    ) -> Result<PacketNumber, PacketDecodeError>
-    where
-        H: crypto::HeaderKey,
-    {
+        header_crypto: &dyn crypto::HeaderKey,
+    ) -> Result<PacketNumber, PacketDecodeError> {
         let packet_length = buf.get_ref().len();
         let pn_offset = buf.position() as usize;
         if packet_length < pn_offset + 4 + header_crypto.sample_size() {
@@ -438,11 +442,12 @@ pub(crate) struct PartialEncode {
 }
 
 impl PartialEncode {
-    pub(crate) fn finish<K, H>(self, buf: &mut [u8], header_crypto: &H, crypto: Option<(u64, &K)>)
-    where
-        K: crypto::PacketKey,
-        H: crypto::HeaderKey,
-    {
+    pub(crate) fn finish(
+        self,
+        buf: &mut [u8],
+        header_crypto: &dyn crypto::HeaderKey,
+        crypto: Option<(u64, &dyn crypto::PacketKey)>,
+    ) {
         let PartialEncode { header_len, pn, .. } = self;
         let (pn_len, write_len) = match pn {
             Some((pn_len, write_len)) => (pn_len, write_len),
@@ -492,7 +497,6 @@ pub(crate) enum PlainHeader {
         version: u32,
     },
     Short {
-        first: u8,
         spin: bool,
         dst_cid: ConnectionId,
     },
@@ -536,9 +540,8 @@ impl PlainHeader {
             }
 
             Ok(PlainHeader::Short {
-                first,
                 spin,
-                dst_cid: ConnectionId::new(&buf.copy_to_bytes(local_cid_len)),
+                dst_cid: ConnectionId::from_buf(buf, local_cid_len),
             })
         } else {
             let version = buf.get::<u32>()?;
@@ -831,13 +834,11 @@ mod tests {
     #[cfg(feature = "rustls")]
     #[test]
     fn header_encoding() {
-        use crate::{
-            crypto::{rustls::TlsSession, PacketKey, Session},
-            Side,
-        };
+        use crate::{crypto::rustls::initial_keys, Side};
+        use rustls::quic::Version;
 
         let dcid = ConnectionId::new(&hex!("06b858ec6f80452b"));
-        let client = TlsSession::initial_keys(&dcid, Side::Client);
+        let client = initial_keys(Version::V1, &dcid, Side::Client);
         let mut buf = Vec::new();
         let header = Header::Initial {
             number: PacketNumber::U8(0),
@@ -851,8 +852,8 @@ mod tests {
         buf.resize(header_len + 16 + client.packet.local.tag_len(), 0);
         encode.finish(
             &mut buf,
-            &client.header.local,
-            Some((0, &client.packet.local)),
+            &*client.header.local,
+            Some((0, &*client.packet.local)),
         );
 
         for byte in &buf {
@@ -862,20 +863,20 @@ mod tests {
         assert_eq!(
             buf[..],
             hex!(
-                "ccff00001d0806b858ec6f80452b00004021b1
-                 b35cedb8ac6906c20c52b09ad2719228310d8f8a51746a75b853060a4e3c0e6e"
+                "c8000000010806b858ec6f80452b00004021be
+                 3ef50807b84191a196f760a6dad1e9d1c430c48952cba0148250c21c0a6a70e1"
             )[..]
         );
 
-        let server = TlsSession::initial_keys(&dcid, Side::Server);
+        let server = initial_keys(Version::V1, &dcid, Side::Server);
         let supported_versions = DEFAULT_SUPPORTED_VERSIONS.to_vec();
         let decode = PartialDecode::new(buf.as_slice().into(), 0, &supported_versions)
             .unwrap()
             .0;
-        let mut packet = decode.finish(Some(&server.header.remote)).unwrap();
+        let mut packet = decode.finish(Some(&*server.header.remote)).unwrap();
         assert_eq!(
             packet.header_data[..],
-            hex!("c0ff00001d0806b858ec6f80452b0000402100")[..]
+            hex!("c0000000010806b858ec6f80452b0000402100")[..]
         );
         server
             .packet
