@@ -14,8 +14,10 @@ use rustls::internal::msgs::enums::AlertDescription;
 use tracing::info;
 
 use super::*;
-use crate::cid_generator::{ConnectionIdGenerator, RandomConnectionIdGenerator};
-use crate::crypto::Session as _;
+use crate::{
+    cid_generator::{ConnectionIdGenerator, RandomConnectionIdGenerator},
+    frame::FrameStruct,
+};
 mod util;
 use util::*;
 
@@ -115,13 +117,46 @@ fn lifecycle() {
     assert_eq!(pair.server.known_cids(), 0);
 }
 
+#[test]
+fn draft_version_compat() {
+    let _guard = subscribe();
+
+    let mut client_config = client_config();
+    client_config.version(0xff00_0020);
+
+    let mut pair = Pair::default();
+    let (client_ch, server_ch) = pair.connect_with(client_config);
+
+    assert_matches!(pair.client_conn_mut(client_ch).poll(), None);
+    assert!(pair.client_conn_mut(client_ch).using_ecn());
+    assert!(pair.server_conn_mut(server_ch).using_ecn());
+
+    const REASON: &[u8] = b"whee";
+    info!("closing");
+    pair.client.connections.get_mut(&client_ch).unwrap().close(
+        pair.time,
+        VarInt(42),
+        REASON.into(),
+    );
+    pair.drive();
+    assert_matches!(pair.server_conn_mut(server_ch).poll(),
+                    Some(Event::ConnectionLost { reason: ConnectionError::ApplicationClosed(
+                        ApplicationClose { error_code: VarInt(42), ref reason }
+                    )}) if reason == REASON);
+    assert_matches!(pair.client_conn_mut(client_ch).poll(), None);
+    assert_eq!(pair.client.known_connections(), 0);
+    assert_eq!(pair.client.known_cids(), 0);
+    assert_eq!(pair.server.known_connections(), 0);
+    assert_eq!(pair.server.known_cids(), 0);
+}
+
 #[test]
 fn stateless_retry() {
     let _guard = subscribe();
     let mut pair = Pair::new(
         Default::default(),
         ServerConfig {
-            use_stateless_retry: true,
+            use_retry: true,
             ..server_config()
         },
     );
@@ -136,7 +171,7 @@ fn server_stateless_reset() {
     rng.fill_bytes(&mut reset_key);
     let reset_key = hmac::Key::new(hmac::HMAC_SHA256, &reset_key);
 
-    let endpoint_config = Arc::new(EndpointConfig::new(reset_key));
+    let endpoint_config = Arc::new(EndpointConfig::new(Arc::new(reset_key)));
 
     let mut pair = Pair::new(endpoint_config.clone(), server_config());
     let (client_ch, _) = pair.connect();
@@ -165,7 +200,7 @@ fn client_stateless_reset() {
     rng.fill_bytes(&mut reset_key);
     let reset_key = hmac::Key::new(hmac::HMAC_SHA256, &reset_key);
 
-    let endpoint_config = Arc::new(EndpointConfig::new(reset_key));
+    let endpoint_config = Arc::new(EndpointConfig::new(Arc::new(reset_key)));
 
     let mut pair = Pair::new(endpoint_config.clone(), server_config());
     let (_, server_ch) = pair.connect();
@@ -318,7 +353,7 @@ fn reject_self_signed_server_cert() {
     let _guard = subscribe();
     let mut pair = Pair::default();
     info!("connecting");
-    let client_ch = pair.begin_connect(ClientConfig::default());
+    let client_ch = pair.begin_connect(client_config_with_certs(vec![]));
     pair.drive();
     assert_matches!(pair.client_conn_mut(client_ch).poll(),
                     Some(Event::ConnectionLost { reason: ConnectionError::TransportError(ref error)})
@@ -328,11 +363,26 @@ fn reject_self_signed_server_cert() {
 #[test]
 fn reject_missing_client_cert() {
     let _guard = subscribe();
-    let mut server_config = server_config();
-    Arc::make_mut(&mut server_config.crypto).set_client_certificate_verifier(
-        rustls::AllowAnyAuthenticatedClient::new(rustls::RootCertStore::empty()),
+
+    let key = rustls::PrivateKey(CERTIFICATE.serialize_private_key_der());
+    let cert = util::CERTIFICATE.serialize_der().unwrap();
+
+    let config = rustls::ServerConfig::builder()
+        .with_safe_default_cipher_suites()
+        .with_safe_default_kx_groups()
+        .with_protocol_versions(&[&rustls::version::TLS13])
+        .unwrap()
+        .with_client_cert_verifier(rustls::server::AllowAnyAuthenticatedClient::new(
+            rustls::RootCertStore::empty(),
+        ))
+        .with_single_cert(vec![rustls::Certificate(cert)], key)
+        .unwrap();
+
+    let mut pair = Pair::new(
+        Default::default(),
+        ServerConfig::with_crypto(Arc::new(config)),
     );
-    let mut pair = Pair::new(Default::default(), server_config);
+
     info!("connecting");
     let client_ch = pair.begin_connect(client_config());
     pair.drive();
@@ -368,17 +418,17 @@ fn congestion() {
     let (client_ch, _) = pair.connect();
 
     const TARGET: u64 = 2048;
-    assert!(pair.client_conn_mut(client_ch).congestion_state() > TARGET);
+    assert!(pair.client_conn_mut(client_ch).congestion_window() > TARGET);
     let s = pair.client_streams(client_ch).open(Dir::Uni).unwrap();
     // Send data without receiving ACKs until the congestion state falls below target
-    while pair.client_conn_mut(client_ch).congestion_state() > TARGET {
+    while pair.client_conn_mut(client_ch).congestion_window() > TARGET {
         let n = pair.client_send(client_ch, s).write(&[42; 1024]).unwrap();
         assert_eq!(n, 1024);
         pair.drive_client();
     }
     // Ensure that the congestion state recovers after receiving the ACKs
     pair.drive();
-    assert!(pair.client_conn_mut(client_ch).congestion_state() >= TARGET);
+    assert!(pair.client_conn_mut(client_ch).congestion_window() >= TARGET);
     pair.client_send(client_ch, s).write(&[42; 1024]).unwrap();
 }
 
@@ -401,7 +451,7 @@ fn zero_rtt_happypath() {
     let mut pair = Pair::new(
         Default::default(),
         ServerConfig {
-            use_stateless_retry: true,
+            use_retry: true,
             ..server_config()
         },
     );
@@ -445,18 +495,16 @@ fn zero_rtt_happypath() {
 #[test]
 fn zero_rtt_rejection() {
     let _guard = subscribe();
-    let mut server_config = server_config();
-    Arc::get_mut(&mut server_config.crypto)
-        .unwrap()
-        .set_protocols(&["foo".into(), "bar".into()]);
+    let mut server_crypto = server_crypto();
+    server_crypto.alpn_protocols = vec!["foo".into(), "bar".into()];
+    let server_config = ServerConfig::with_crypto(Arc::new(server_crypto));
     let mut pair = Pair::new(Arc::new(EndpointConfig::default()), server_config);
-    let mut client_config = client_config();
-    Arc::get_mut(&mut client_config.crypto)
-        .unwrap()
-        .set_protocols(&["foo".into()]);
+    let mut client_crypto = client_crypto();
+    client_crypto.alpn_protocols = vec!["foo".into()];
+    let client_config = ClientConfig::new(Arc::new(client_crypto.clone()));
 
     // Establish normal connection
-    let client_ch = pair.begin_connect(client_config.clone());
+    let client_ch = pair.begin_connect(client_config);
     pair.drive();
     let server_ch = pair.server.assert_accept();
     assert_matches!(
@@ -483,9 +531,8 @@ fn zero_rtt_rejection() {
     pair.server.connections.clear();
 
     // Changing protocols invalidates 0-RTT
-    Arc::get_mut(&mut client_config.crypto)
-        .unwrap()
-        .set_protocols(&["bar".into()]);
+    client_crypto.alpn_protocols = vec!["bar".into()];
+    let client_config = ClientConfig::new(Arc::new(client_crypto));
     info!("resuming session");
     let client_ch = pair.begin_connect(client_config);
     assert!(pair.client_conn_mut(client_ch).has_0rtt());
@@ -517,15 +564,13 @@ fn zero_rtt_rejection() {
 #[test]
 fn alpn_success() {
     let _guard = subscribe();
-    let mut server_config = server_config();
-    Arc::get_mut(&mut server_config.crypto)
-        .unwrap()
-        .set_protocols(&["foo".into(), "bar".into(), "baz".into()]);
+    let mut server_crypto = server_crypto();
+    server_crypto.alpn_protocols = vec!["foo".into(), "bar".into(), "baz".into()];
+    let server_config = ServerConfig::with_crypto(Arc::new(server_crypto));
     let mut pair = Pair::new(Arc::new(EndpointConfig::default()), server_config);
-    let mut client_config = client_config();
-    Arc::get_mut(&mut client_config.crypto)
-        .unwrap()
-        .set_protocols(&["bar".into(), "quux".into(), "corge".into()]);
+    let mut client_crypto = client_crypto();
+    client_crypto.alpn_protocols = vec!["bar".into(), "quux".into(), "corge".into()];
+    let client_config = ClientConfig::new(Arc::new(client_crypto));
 
     // Establish normal connection
     let client_ch = pair.begin_connect(client_config);
@@ -544,6 +589,8 @@ fn alpn_success() {
         .client_conn_mut(client_ch)
         .crypto_session()
         .handshake_data()
+        .unwrap()
+        .downcast::<crate::crypto::rustls::HandshakeData>()
         .unwrap();
     assert_eq!(hd.protocol.unwrap(), &b"bar"[..]);
 }
@@ -552,10 +599,10 @@ fn alpn_success() {
 fn server_alpn_unset() {
     let _guard = subscribe();
     let mut pair = Pair::new(Arc::new(EndpointConfig::default()), server_config());
-    let mut client_config = client_config();
-    Arc::get_mut(&mut client_config.crypto)
-        .unwrap()
-        .set_protocols(&["foo".into()]);
+
+    let mut client_crypto = client_crypto();
+    client_crypto.alpn_protocols = vec!["foo".into()];
+    let client_config = ClientConfig::new(Arc::new(client_crypto));
 
     let client_ch = pair.begin_connect(client_config);
     pair.drive();
@@ -568,10 +615,9 @@ fn server_alpn_unset() {
 #[test]
 fn client_alpn_unset() {
     let _guard = subscribe();
-    let mut server_config = server_config();
-    Arc::get_mut(&mut server_config.crypto)
-        .unwrap()
-        .set_protocols(&["foo".into(), "bar".into(), "baz".into()]);
+    let mut server_crypto = server_crypto();
+    server_crypto.alpn_protocols = vec!["foo".into(), "bar".into(), "baz".into()];
+    let server_config = ServerConfig::with_crypto(Arc::new(server_crypto));
     let mut pair = Pair::new(Arc::new(EndpointConfig::default()), server_config);
 
     let client_ch = pair.begin_connect(client_config());
@@ -584,15 +630,14 @@ fn client_alpn_unset() {
 
 #[test]
 fn alpn_mismatch() {
-    let mut server_config = server_config();
-    Arc::get_mut(&mut server_config.crypto)
-        .unwrap()
-        .set_protocols(&["foo".into(), "bar".into(), "baz".into()]);
+    let mut server_crypto = server_crypto();
+    server_crypto.alpn_protocols = vec!["foo".into(), "bar".into(), "baz".into()];
+    let server_config = ServerConfig::with_crypto(Arc::new(server_crypto));
     let mut pair = Pair::new(Arc::new(EndpointConfig::default()), server_config);
-    let mut client_config = client_config();
-    Arc::get_mut(&mut client_config.crypto)
-        .unwrap()
-        .set_protocols(&["quux".into(), "corge".into()]);
+
+    let mut client_crypto = client_crypto();
+    client_crypto.alpn_protocols = vec!["quux".into(), "corge".into()];
+    let client_config = ClientConfig::new(Arc::new(client_crypto));
 
     let client_ch = pair.begin_connect(client_config);
     pair.drive();
@@ -868,10 +913,10 @@ fn instant_close_2() {
 #[test]
 fn idle_timeout() {
     let _guard = subscribe();
-    const IDLE_TIMEOUT: Duration = Duration::from_millis(10);
+    const IDLE_TIMEOUT: u64 = 100;
     let server = ServerConfig {
         transport: Arc::new(TransportConfig {
-            max_idle_timeout: Some(IDLE_TIMEOUT),
+            max_idle_timeout: Some(VarInt(IDLE_TIMEOUT)),
             ..TransportConfig::default()
         }),
         ..server_config()
@@ -892,7 +937,7 @@ fn idle_timeout() {
         pair.client.inbound.clear(); // Simulate total S->C packet loss
     }
 
-    assert!(pair.time - start < 2 * IDLE_TIMEOUT);
+    assert!(pair.time - start < Duration::from_millis(2 * IDLE_TIMEOUT));
     assert_matches!(
         pair.client_conn_mut(client_ch).poll(),
         Some(Event::ConnectionLost {
@@ -907,6 +952,30 @@ fn idle_timeout() {
     );
 }
 
+#[test]
+fn connection_close_sends_acks() {
+    let _guard = subscribe();
+    let mut pair = Pair::default();
+    let (client_ch, _server_ch) = pair.connect();
+
+    let client_acks = pair.client_conn_mut(client_ch).stats().frame_rx.acks;
+
+    pair.client_conn_mut(client_ch).ping();
+    pair.drive_client();
+
+    let time = pair.time;
+    pair.server_conn_mut(client_ch)
+        .close(time, VarInt(42), Bytes::new());
+
+    pair.drive();
+
+    let client_acks_2 = pair.client_conn_mut(client_ch).stats().frame_rx.acks;
+    assert!(
+        client_acks_2 > client_acks,
+        "Connection close should send pending ACKs"
+    );
+}
+
 #[test]
 fn concurrent_connections_full() {
     let _guard = subscribe();
@@ -962,6 +1031,13 @@ fn migration() {
         CLIENT_PORTS.lock().unwrap().next().unwrap(),
     );
     pair.client_conn_mut(client_ch).ping();
+
+    // Assert that just receiving the ping message is accounted into the servers
+    // anti-amplification budget
+    pair.drive_client();
+    pair.drive_server();
+    assert_ne!(pair.server_conn_mut(server_ch).total_recvd(), 0);
+
     pair.drive();
     assert_matches!(pair.client_conn_mut(client_ch).poll(), None);
     assert_eq!(
@@ -1186,11 +1262,11 @@ fn zero_length_cid() {
 #[test]
 fn keep_alive() {
     let _guard = subscribe();
-    const IDLE_TIMEOUT: Duration = Duration::from_secs(10);
+    const IDLE_TIMEOUT: u64 = 10;
     let server = ServerConfig {
         transport: Arc::new(TransportConfig {
-            keep_alive_interval: Some(IDLE_TIMEOUT / 2),
-            max_idle_timeout: Some(IDLE_TIMEOUT),
+            keep_alive_interval: Some(Duration::from_millis(IDLE_TIMEOUT / 2)),
+            max_idle_timeout: Some(VarInt(IDLE_TIMEOUT)),
             ..TransportConfig::default()
         }),
         ..server_config()
@@ -1198,7 +1274,7 @@ fn keep_alive() {
     let mut pair = Pair::new(Default::default(), server);
     let (client_ch, server_ch) = pair.connect();
     // Run a good while longer than the idle timeout
-    let end = pair.time + 20 * IDLE_TIMEOUT;
+    let end = pair.time + Duration::from_millis(20 * IDLE_TIMEOUT);
     while pair.time < end {
         if !pair.step() {
             if let Some(time) = min_opt(pair.client.next_wakeup(), pair.server.next_wakeup()) {
@@ -1439,10 +1515,10 @@ fn congested_tail_loss() {
     let (client_ch, _) = pair.connect();
 
     const TARGET: u64 = 2048;
-    assert!(pair.client_conn_mut(client_ch).congestion_state() > TARGET);
+    assert!(pair.client_conn_mut(client_ch).congestion_window() > TARGET);
     let s = pair.client_streams(client_ch).open(Dir::Uni).unwrap();
     // Send data without receiving ACKs until the congestion state falls below target
-    while pair.client_conn_mut(client_ch).congestion_state() > TARGET {
+    while pair.client_conn_mut(client_ch).congestion_window() > TARGET {
         let n = pair.client_send(client_ch, s).write(&[42; 1024]).unwrap();
         assert_eq!(n, 1024);
         pair.drive_client();
@@ -1452,7 +1528,7 @@ fn congested_tail_loss() {
     // Ensure that the congestion state recovers after retransmits occur and are ACKed
     info!("recovering");
     pair.drive();
-    assert!(pair.client_conn_mut(client_ch).congestion_state() > TARGET);
+    assert!(pair.client_conn_mut(client_ch).congestion_window() > TARGET);
     pair.client_send(client_ch, s).write(&[42; 1024]).unwrap();
 }
 
@@ -1489,9 +1565,9 @@ fn datagram_recv_buffer_overflow() {
     let mut pair = Pair::new(Default::default(), server);
     let (client_ch, server_ch) = pair.connect();
     assert_matches!(pair.server_conn_mut(server_ch).poll(), None);
-    assert_matches!(
+    assert_eq!(
         pair.client_conn_mut(client_ch).datagrams().max_size(),
-        Some(WINDOW)
+        Some(WINDOW - Datagram::SIZE_BOUND)
     );
 
     const DATA1: &[u8] = &[0xAB; (WINDOW / 3) + 1];
@@ -1540,18 +1616,17 @@ fn datagram_unsupported() {
 #[test]
 fn large_initial() {
     let _guard = subscribe();
-    let mut server_config = server_config();
-    Arc::get_mut(&mut server_config.crypto)
-        .unwrap()
-        .set_protocols(&[vec![0, 0, 0, 42]]);
+    let mut server_crypto = server_crypto();
+    server_crypto.alpn_protocols = vec![vec![0, 0, 0, 42]];
+    let server_config = ServerConfig::with_crypto(Arc::new(server_crypto));
+
     let mut pair = Pair::new(Arc::new(EndpointConfig::default()), server_config);
-    let mut cfg = client_config();
+    let mut client_crypto = client_crypto();
     let protocols = (0..1000u32)
         .map(|x| x.to_be_bytes().to_vec())
         .collect::<Vec<_>>();
-    Arc::get_mut(&mut cfg.crypto)
-        .unwrap()
-        .set_protocols(&protocols);
+    client_crypto.alpn_protocols = protocols;
+    let cfg = ClientConfig::new(Arc::new(client_crypto));
     let client_ch = pair.begin_connect(cfg);
     pair.drive();
     let server_ch = pair.server.assert_accept();
@@ -1725,3 +1800,78 @@ fn repeated_request_response() {
         let _ = chunks.finalize();
     }
 }
+
+/// Ensures that the client sends an anti-deadlock probe after an incomplete server's first flight
+#[test]
+fn handshake_anti_deadlock_probe() {
+    let _guard = subscribe();
+
+    let (cert, key) = big_cert_and_key();
+    let server = server_config_with_cert(cert.clone(), key);
+    let client = client_config_with_certs(vec![cert]);
+    let mut pair = Pair::new(Default::default(), server);
+
+    let client_ch = pair.begin_connect(client);
+    // Client sends initial
+    pair.drive_client();
+    // Server sends first flight, gets blocked on anti-amplification
+    pair.drive_server();
+    // Client acks...
+    pair.drive_client();
+    // ...but it's lost, so the server doesn't get anti-amplification credit from it
+    pair.server.inbound.clear();
+    // Client sends an anti-deadlock probe, and the handshake completes as usual.
+    pair.drive();
+    assert_matches!(
+        pair.client_conn_mut(client_ch).poll(),
+        Some(Event::HandshakeDataReady)
+    );
+    assert_matches!(
+        pair.client_conn_mut(client_ch).poll(),
+        Some(Event::Connected { .. })
+    );
+}
+
+/// Ensures that the server can respond with 3 initial packets during the handshake
+/// before the anti-amplification limit kicks in when MTUs are similar.
+#[test]
+fn server_can_send_3_inital_packets() {
+    let _guard = subscribe();
+
+    let (cert, key) = big_cert_and_key();
+    let server = server_config_with_cert(cert.clone(), key);
+    let client = client_config_with_certs(vec![cert]);
+    let mut pair = Pair::new(Default::default(), server);
+
+    let client_ch = pair.begin_connect(client);
+    // Client sends initial
+    pair.drive_client();
+    // Server sends first flight, gets blocked on anti-amplification
+    pair.drive_server();
+    // Server should have queued 3 packets at this time
+    assert_eq!(pair.client.inbound.len(), 3);
+
+    pair.drive();
+    assert_matches!(
+        pair.client_conn_mut(client_ch).poll(),
+        Some(Event::HandshakeDataReady)
+    );
+    assert_matches!(
+        pair.client_conn_mut(client_ch).poll(),
+        Some(Event::Connected { .. })
+    );
+}
+
+/// Generate a big fat certificate that can't fit inside the initial anti-amplification limit
+fn big_cert_and_key() -> (rustls::Certificate, rustls::PrivateKey) {
+    let cert = rcgen::generate_simple_self_signed(
+        Some("localhost".into())
+            .into_iter()
+            .chain((0..1000).map(|x| format!("foo_{}", x)))
+            .collect::<Vec<_>>(),
+    )
+    .unwrap();
+    let key = rustls::PrivateKey(cert.serialize_private_key_der());
+    let cert = rustls::Certificate(cert.serialize_der().unwrap());
+    (cert, key)
+}
